@@ -56,7 +56,7 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
     };
 
     info!(
-        "Found predecessors for {} syntax nodes (hashmap key: {} bytes, value: {} bytes), with {} left on heap.",
+        "Found predecessors for {} vertices (hashmap key: {} bytes, value: {} bytes), with {} left on heap.",
         predecessors.len(),
         std::mem::size_of::<Vertex>(),
         std::mem::size_of::<PredecessorInfo>(),
@@ -83,12 +83,51 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
     info!(
         "Initial {} items on path: {:#?}",
         print_length,
-        route.iter().map(|x| x.0).take(print_length).collect_vec()
+        route
+            .iter()
+            .map(|x| {
+                format!(
+                    "{:10} {:10} --- {:?}",
+                    x.1.lhs_syntax
+                        .map(Syntax::dbg_content)
+                        .unwrap_or_else(|| "None".into()),
+                    x.1.rhs_syntax
+                        .map(Syntax::dbg_content)
+                        .unwrap_or_else(|| "None".into()),
+                    x.0,
+                )
+            })
+            .take(print_length)
+            .collect_vec()
     );
     route
 }
 
+fn node_count(root: Option<&Syntax>) -> u32 {
+    let mut node = root;
+    let mut count = 0;
+    while let Some(current_node) = node {
+        let current_count = match current_node {
+            Syntax::List {
+                num_descendants, ..
+            } => *num_descendants,
+            Syntax::Atom { .. } => 1,
+        };
+        count += current_count;
+
+        node = current_node.next_sibling();
+    }
+
+    count
+}
+
 pub fn mark_syntax<'a>(lhs_syntax: Option<&'a Syntax<'a>>, rhs_syntax: Option<&'a Syntax<'a>>) {
+    info!(
+        "LHS nodes: {}, RHS nodes: {}",
+        node_count(lhs_syntax),
+        node_count(rhs_syntax)
+    );
+
     let start = Vertex::new(lhs_syntax, rhs_syntax);
     let route = shortest_path(start);
     mark_route(&route);
@@ -135,7 +174,12 @@ mod tests {
         let route = shortest_path(start);
 
         let actions = route.iter().map(|(action, _)| *action).collect_vec();
-        assert_eq!(actions, vec![UnchangedNode]);
+        assert_eq!(
+            actions,
+            vec![UnchangedNode {
+                depth_difference: 0
+            }]
+        );
     }
 
     #[test]
@@ -173,7 +217,9 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                EnterUnchangedDelimiter,
+                EnterUnchangedDelimiter {
+                    depth_difference: 0
+                },
                 NovelAtomLHS { contiguous: false },
                 ExitDelimiterBoth,
             ]
@@ -213,7 +259,9 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                EnterUnchangedDelimiter,
+                EnterUnchangedDelimiter {
+                    depth_difference: 0
+                },
                 NovelAtomRHS { contiguous: false },
                 NovelAtomRHS { contiguous: false },
                 ExitDelimiterBoth,
@@ -259,10 +307,14 @@ mod tests {
             vec![
                 EnterNovelDelimiterRHS { contiguous: false },
                 EnterNovelDelimiterLHS { contiguous: false },
-                UnchangedNode,
-                UnchangedNode,
-                ExitDelimiterLHS,
+                UnchangedNode {
+                    depth_difference: 0
+                },
+                UnchangedNode {
+                    depth_difference: 0
+                },
                 ExitDelimiterRHS,
+                ExitDelimiterLHS,
             ],
         );
     }
@@ -292,7 +344,9 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                UnchangedNode,
+                UnchangedNode {
+                    depth_difference: 0
+                },
                 NovelAtomLHS { contiguous: false },
                 NovelAtomLHS { contiguous: true },
             ]
