@@ -150,6 +150,8 @@ fn changed_parts<'a>(
 
 // TODO: Prefer src/opposite_src nomenclature as this function is called from both sides.
 pub fn change_positions(lhs_src: &str, rhs_src: &str) -> Vec<MatchedPos> {
+    // TODO: If either side is "", don't split each line by words
+    // pointlessly. This is common for file additions/removals.
     let lhs_nlp = NewlinePositions::from(lhs_src);
     let rhs_nlp = NewlinePositions::from(rhs_src);
 
@@ -184,7 +186,7 @@ pub fn change_positions(lhs_src: &str, rhs_src: &str) -> Vec<MatchedPos> {
                 for diff_res in diff::slice(&split_words(&lhs_part), &split_words(&rhs_part)) {
                     match diff_res {
                         diff::Result::Left(lhs_word) => {
-                            if lhs_word != "\n" {
+                            if *lhs_word != "\n" {
                                 let lhs_pos =
                                     lhs_nlp.from_offsets(lhs_offset, lhs_offset + lhs_word.len());
                                 res.push(MatchedPos {
@@ -198,7 +200,7 @@ pub fn change_positions(lhs_src: &str, rhs_src: &str) -> Vec<MatchedPos> {
                             lhs_offset += lhs_word.len();
                         }
                         diff::Result::Both(lhs_word, rhs_word) => {
-                            if lhs_word != "\n" {
+                            if *lhs_word != "\n" {
                                 let lhs_pos =
                                     lhs_nlp.from_offsets(lhs_offset, lhs_offset + lhs_word.len());
                                 let rhs_pos =
