@@ -151,16 +151,6 @@ pub fn codepoint_len(s: &str) -> usize {
     s.chars().count()
 }
 
-/// The first `len` codepoints of `s`. This is safer than slicing by
-/// bytes, which panics if the byte isn't on a codepoint boundary.
-pub fn substring_by_codepoint(s: &str, start: usize, end: usize) -> &str {
-    let byte_start = s.char_indices().nth(start).unwrap().0;
-    match s.char_indices().nth(end) {
-        Some(byte_end) => &s[byte_start..byte_end.0],
-        None => &s[byte_start..],
-    }
-}
-
 pub trait MaxLine {
     fn max_line(&self) -> LineNumber;
 }
