@@ -51,7 +51,6 @@ use style::BackgroundColor;
 use summary::{DiffResult, FileContent};
 use syntax::init_next_prev;
 use typed_arena::Arena;
-use unchanged::skip_unchanged;
 use walkdir::WalkDir;
 
 use crate::{
@@ -249,14 +248,17 @@ fn diff_file_content(
 
             init_all_info(&lhs, &rhs);
 
-            let (possibly_changed_lhs, possibly_changed_rhs) = skip_unchanged(&lhs, &rhs);
+            let possibly_changed = if env::var("DFT_DBG_KEEP_UNCHANGED").is_ok() {
+                vec![(lhs.clone(), rhs.clone())]
+            } else {
+                unchanged::mark_unchanged(&lhs, &rhs)
+            };
 
-            let possibly_changed_num =
-                num_nodes(&possibly_changed_lhs) + num_nodes(&possibly_changed_rhs);
-            if possibly_changed_num > node_limit {
+            let possibly_changed_max = max_num_nodes(&possibly_changed);
+            if possibly_changed_max > node_limit {
                 info!(
                     "Found {} nodes, exceeding the limit {}",
-                    possibly_changed_num, node_limit
+                    possibly_changed_max, node_limit
                 );
 
                 let lhs_positions = line_parser::change_positions(&lhs_src, &rhs_src);
@@ -267,16 +269,18 @@ fn diff_file_content(
                     rhs_positions,
                 )
             } else {
-                init_next_prev(&possibly_changed_lhs);
-                init_next_prev(&possibly_changed_rhs);
+                for (lhs_section_nodes, rhs_section_nodes) in possibly_changed {
+                    init_next_prev(&lhs_section_nodes);
+                    init_next_prev(&rhs_section_nodes);
 
-                mark_syntax(
-                    possibly_changed_lhs.get(0).copied(),
-                    possibly_changed_rhs.get(0).copied(),
-                );
+                    mark_syntax(
+                        lhs_section_nodes.get(0).copied(),
+                        rhs_section_nodes.get(0).copied(),
+                    );
 
-                fix_all_sliders(&possibly_changed_lhs);
-                fix_all_sliders(&possibly_changed_rhs);
+                    fix_all_sliders(&lhs_section_nodes);
+                    fix_all_sliders(&rhs_section_nodes);
+                }
 
                 let lhs_positions = syntax::change_positions(&lhs);
                 let rhs_positions = syntax::change_positions(&rhs);
@@ -443,6 +447,14 @@ fn num_nodes(roots: &[&syntax::Syntax]) -> u32 {
         .sum()
 }
 
+fn max_num_nodes(roots_vec: &[(Vec<&syntax::Syntax>, Vec<&syntax::Syntax>)]) -> u32 {
+    roots_vec
+        .iter()
+        .map(|(lhs, rhs)| num_nodes(lhs) + num_nodes(rhs))
+        .max()
+        .unwrap_or(0)
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
