@@ -1,6 +1,6 @@
 //! Side-by-side (two column) display of diffs.
 
-use colored::{Color, Colorize};
+use owo_colors::{OwoColorize, Style};
 use std::{
     cmp::max,
     collections::{HashMap, HashSet},
@@ -11,7 +11,7 @@ use crate::{
     hunks::{matched_lines_for_hunk, Hunk},
     lines::{codepoint_len, format_line_num, LineNumber},
     positions::SingleLineSpan,
-    style::{self, apply_colors, color_positions, split_and_apply, BackgroundColor, Style},
+    style::{self, apply_colors, color_positions, novel_style, split_and_apply, BackgroundColor},
     syntax::{zip_pad_shorter, MatchedPos},
 };
 
@@ -70,25 +70,31 @@ fn display_single_column(
     lang_name: &str,
     src: &str,
     is_lhs: bool,
+    use_color: bool,
     background: BackgroundColor,
 ) -> String {
     let column_width = format_line_num(src.lines().count().into()).len();
 
     let mut result = String::with_capacity(src.len());
-    result.push_str(&style::header(display_path, 1, 1, lang_name, background));
+    result.push_str(&style::header(
+        display_path,
+        1,
+        1,
+        lang_name,
+        use_color,
+        background,
+    ));
     result.push('\n');
 
-    let color = match (is_lhs, background) {
-        (true, BackgroundColor::Dark) => Color::BrightRed,
-        (true, BackgroundColor::Light) => Color::Red,
-        (false, BackgroundColor::Dark) => Color::BrightGreen,
-        (false, BackgroundColor::Light) => Color::Green,
-    };
+    let mut style = Style::new();
+    if use_color {
+        style = novel_style(Style::new(), is_lhs, background);
+    }
 
     for (i, line) in src.lines().enumerate() {
         result.push_str(
             &format_line_num_padded(i.into(), column_width)
-                .color(color)
+                .style(style)
                 .to_string(),
         );
         result.push_str(line);
@@ -102,6 +108,7 @@ fn display_line_nums(
     lhs_line_num: Option<LineNumber>,
     rhs_line_num: Option<LineNumber>,
     source_dims: &SourceDimensions,
+    use_color: bool,
     background: BackgroundColor,
     lhs_has_novel: bool,
     rhs_has_novel: bool,
@@ -111,13 +118,13 @@ fn display_line_nums(
     let display_lhs_line_num: String = match lhs_line_num {
         Some(line_num) => {
             let s = format_line_num_padded(line_num, source_dims.lhs_line_nums_width);
-            if lhs_has_novel {
+            if lhs_has_novel && use_color {
                 // TODO: factor out applying colours to line numbers.
-                match background {
-                    BackgroundColor::Dark => s.bright_red(),
-                    BackgroundColor::Light => s.red(),
+                if background.is_dark() {
+                    s.bright_red().to_string()
+                } else {
+                    s.red().to_string()
                 }
-                .to_string()
             } else {
                 s
             }
@@ -131,12 +138,12 @@ fn display_line_nums(
     let display_rhs_line_num: String = match rhs_line_num {
         Some(line_num) => {
             let s = format_line_num_padded(line_num, source_dims.rhs_line_nums_width);
-            if rhs_has_novel {
-                match background {
-                    BackgroundColor::Dark => s.bright_green(),
-                    BackgroundColor::Light => s.green(),
+            if rhs_has_novel && use_color {
+                if background.is_dark() {
+                    s.bright_green().to_string()
+                } else {
+                    s.green().to_string()
                 }
-                .to_string()
             } else {
                 s
             }
@@ -278,6 +285,7 @@ fn highlight_as_novel(
 pub fn print(
     hunks: &[Hunk],
     display_width: usize,
+    use_color: bool,
     background: BackgroundColor,
     display_path: &str,
     lang_name: &str,
@@ -286,24 +294,51 @@ pub fn print(
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) {
-    let lhs_colored_src = apply_colors(lhs_src, true, background, lhs_mps);
-    let rhs_colored_src = apply_colors(rhs_src, false, background, rhs_mps);
+    let (lhs_colored_src, rhs_colored_src) = if use_color {
+        (
+            apply_colors(lhs_src, true, background, lhs_mps),
+            apply_colors(rhs_src, false, background, rhs_mps),
+        )
+    } else {
+        (lhs_src.to_string(), rhs_src.to_string())
+    };
 
     if lhs_src.is_empty() {
         println!(
             "{}",
-            display_single_column(display_path, lang_name, &rhs_colored_src, false, background)
+            display_single_column(
+                display_path,
+                lang_name,
+                &rhs_colored_src,
+                false,
+                use_color,
+                background
+            )
         );
+        return;
     }
     if rhs_src.is_empty() {
         println!(
             "{}",
-            display_single_column(display_path, lang_name, &lhs_colored_src, true, background)
+            display_single_column(
+                display_path,
+                lang_name,
+                &lhs_colored_src,
+                true,
+                use_color,
+                background
+            )
         );
+        return;
     }
 
     // TODO: this is largely duplicating the `apply_colors` logic.
-    let (lhs_highlights, rhs_highlights) = highlight_positions(background, lhs_mps, rhs_mps);
+    let (lhs_highlights, rhs_highlights) = if use_color {
+        highlight_positions(background, lhs_mps, rhs_mps)
+    } else {
+        (HashMap::new(), HashMap::new())
+    };
+
     let lhs_lines = split_on_newlines(lhs_src);
     let rhs_lines = split_on_newlines(rhs_src);
     let lhs_colored_lines = split_on_newlines(&lhs_colored_src);
@@ -319,7 +354,14 @@ pub fn print(
     for (i, hunk) in hunks.iter().enumerate() {
         println!(
             "{}",
-            style::header(display_path, i + 1, hunks.len(), lang_name, background)
+            style::header(
+                display_path,
+                i + 1,
+                hunks.len(),
+                lang_name,
+                use_color,
+                background
+            )
         );
 
         let aligned_lines = matched_lines_for_hunk(&matched_lines, hunk);
@@ -347,6 +389,7 @@ pub fn print(
                 lhs_line_num,
                 rhs_line_num,
                 &source_dims,
+                use_color,
                 background,
                 lhs_line_novel,
                 rhs_line_novel,
@@ -354,7 +397,7 @@ pub fn print(
                 prev_rhs_line_num,
             );
 
-            if no_lhs_changes {
+            if no_lhs_changes && !std::env::var("DFT_SHOW_BOTH").is_ok() {
                 match rhs_line_num {
                     Some(rhs_line_num) => {
                         let rhs_line = &rhs_colored_lines[rhs_line_num.0];
@@ -374,7 +417,7 @@ pub fn print(
                         println!("{}{}", display_rhs_line_num, display_rhs_line_num);
                     }
                 }
-            } else if no_rhs_changes {
+            } else if no_rhs_changes && !std::env::var("DFT_SHOW_BOTH").is_ok() {
                 match lhs_line_num {
                     Some(lhs_line_num) => {
                         let lhs_line = &lhs_colored_lines[lhs_line_num.0];
@@ -396,6 +439,7 @@ pub fn print(
                     Some(lhs_line_num) => split_and_apply(
                         lhs_lines[lhs_line_num.0],
                         source_dims.lhs_content_width,
+                        use_color,
                         lhs_highlights.get(&lhs_line_num).unwrap_or(&vec![]),
                     ),
                     None => vec![" ".repeat(source_dims.lhs_content_width)],
@@ -404,6 +448,7 @@ pub fn print(
                     Some(rhs_line_num) => split_and_apply(
                         rhs_lines[rhs_line_num.0],
                         source_dims.rhs_content_width,
+                        use_color,
                         rhs_highlights.get(&rhs_line_num).unwrap_or(&vec![]),
                     ),
                     None => vec!["".into()],
@@ -427,11 +472,11 @@ pub fn print(
                         );
                         if let Some(line_num) = lhs_line_num {
                             if lhs_lines_with_novel.contains(&line_num) {
-                                s = match background {
-                                    BackgroundColor::Dark => s.bright_red(),
-                                    BackgroundColor::Light => s.red(),
-                                }
-                                .to_string();
+                                s = if background.is_dark() {
+                                    s.bright_red().to_string()
+                                } else {
+                                    s.red().to_string()
+                                };
                             }
                         }
                         s
@@ -447,11 +492,11 @@ pub fn print(
                         );
                         if let Some(line_num) = rhs_line_num {
                             if rhs_lines_with_novel.contains(&line_num) {
-                                s = match background {
-                                    BackgroundColor::Dark => s.bright_green(),
-                                    BackgroundColor::Light => s.green(),
-                                }
-                                .to_string();
+                                s = if background.is_dark() {
+                                    s.bright_green().to_string()
+                                } else {
+                                    s.green().to_string()
+                                };
                             }
                         }
                         s
@@ -537,6 +582,7 @@ mod tests {
             "Python",
             "print(123)\n",
             false,
+            true,
             BackgroundColor::Dark,
         );
         assert!(res.len() > 10);
@@ -608,6 +654,7 @@ mod tests {
         print(
             &hunks,
             80,
+            true,
             BackgroundColor::Dark,
             "foo.el",
             "Emacs Lisp",
