@@ -193,6 +193,17 @@ impl<'a> Syntax<'a> {
         close_content: &str,
         close_position: Vec<SingleLineSpan>,
     ) -> &'a Syntax<'a> {
+        // Skip empty atoms: they aren't displayed, so there's no
+        // point making our syntax tree bigger. These occur when we're
+        // parsing incomplete or malformed programs.
+        let children = children
+            .into_iter()
+            .filter(|n| match n {
+                List { .. } => true,
+                Atom { content, .. } => !content.is_empty(),
+            })
+            .collect::<Vec<_>>();
+
         // Don't bother creating a list if we have no open/close and
         // there's only one child. This occurs in small files with
         // thorough tree-sitter parsers: you get parse trees like:
@@ -247,6 +258,10 @@ impl<'a> Syntax<'a> {
         }
     }
 
+    pub fn parent(&self) -> Option<&'a Syntax<'a>> {
+        self.info().parent.get()
+    }
+
     pub fn next_sibling(&self) -> Option<&'a Syntax<'a>> {
         self.info().next_sibling.get()
     }
@@ -272,6 +287,17 @@ impl<'a> Syntax<'a> {
         self.info().num_ancestors.get()
     }
 
+    pub fn dbg_content(&self) -> String {
+        match self {
+            List {
+                open_content,
+                close_content,
+                ..
+            } => format!("{} ... {}", open_content, close_content),
+            Atom { content, .. } => content.into(),
+        }
+    }
+
     pub fn first_line(&self) -> Option<LineNumber> {
         let position = match self {
             List { open_position, .. } => open_position,
@@ -327,6 +353,8 @@ pub fn init_all_info<'a>(lhs_roots: &[&'a Syntax<'a>], rhs_roots: &[&'a Syntax<'
 }
 
 pub fn init_info<'a>(lhs_roots: &[&'a Syntax<'a>], rhs_roots: &[&'a Syntax<'a>]) {
+    // TODO: start unique IDs at 1. This helps with hashing and makes
+    // it obvious if we've forgotten to init the syntax info.
     let mut id = 0;
     init_info_single(lhs_roots, &mut id);
     init_info_single(rhs_roots, &mut id);
@@ -336,13 +364,7 @@ pub fn init_info<'a>(lhs_roots: &[&'a Syntax<'a>], rhs_roots: &[&'a Syntax<'a>])
     set_content_id(rhs_roots, &mut existing);
 }
 
-type ContentKey = (
-    Option<String>,
-    Option<String>,
-    Vec<u32>,
-    bool,
-    Option<AtomKind>,
-);
+type ContentKey = (Option<String>, Option<String>, Vec<u32>, bool, bool);
 
 fn set_content_id(nodes: &[&Syntax], existing: &mut HashMap<ContentKey, u32>) {
     for node in nodes {
@@ -364,7 +386,7 @@ fn set_content_id(nodes: &[&Syntax], existing: &mut HashMap<ContentKey, u32>) {
                     Some(close_content.clone()),
                     children_content_ids,
                     true,
-                    None,
+                    true,
                 )
             }
             Atom {
@@ -372,18 +394,18 @@ fn set_content_id(nodes: &[&Syntax], existing: &mut HashMap<ContentKey, u32>) {
                 kind: highlight,
                 ..
             } => {
-                let clean_content =
-                    if *highlight == AtomKind::Comment && content.lines().count() > 1 {
-                        content
-                            .lines()
-                            .map(|l| l.trim_start())
-                            .collect::<Vec<_>>()
-                            .join("\n")
-                            .to_string()
-                    } else {
-                        content.clone()
-                    };
-                (Some(clean_content), None, vec![], false, Some(*highlight))
+                let is_comment = *highlight == AtomKind::Comment;
+                let clean_content = if is_comment && content.lines().count() > 1 {
+                    content
+                        .lines()
+                        .map(|l| l.trim_start())
+                        .collect::<Vec<_>>()
+                        .join("\n")
+                        .to_string()
+                } else {
+                    content.clone()
+                };
+                (Some(clean_content), None, vec![], false, is_comment)
             }
         };
 
@@ -833,6 +855,8 @@ mod tests {
     use super::*;
     use pretty_assertions::assert_eq;
 
+    /// Consider comment atoms as distinct to other atoms even if the
+    /// content matches otherwise.
     #[test]
     fn test_comment_and_atom_differ() {
         let pos = vec![SingleLineSpan {
@@ -850,6 +874,26 @@ mod tests {
         assert_ne!(comment, atom);
     }
 
+    /// Ignore the syntax highighting kind when comparing
+    /// atoms. Sometimes changing delimiter wrapping can change
+    /// whether a parser thinks that a node is e.g. a type.
+    #[test]
+    fn test_atom_equality_ignores_highlighting() {
+        let pos = vec![SingleLineSpan {
+            line: 0.into(),
+            start_col: 2,
+            end_col: 3,
+        }];
+
+        let arena = Arena::new();
+
+        let type_atom = Syntax::new_atom(&arena, pos.clone(), "foo", AtomKind::Type);
+        let atom = Syntax::new_atom(&arena, pos, "foo", AtomKind::Normal);
+        init_all_info(&[type_atom], &[atom]);
+
+        assert_eq!(type_atom, atom);
+    }
+
     #[test]
     fn test_flatten_trivial_list() {
         let pos = vec![SingleLineSpan {
@@ -866,6 +910,27 @@ mod tests {
         assert!(matches!(trivial_list, Atom { .. }));
     }
 
+    #[test]
+    fn test_ignore_empty_atoms() {
+        let pos = vec![SingleLineSpan {
+            line: 0.into(),
+            start_col: 2,
+            end_col: 2,
+        }];
+
+        let arena = Arena::new();
+        let atom = Syntax::new_atom(&arena, pos, "", AtomKind::Normal);
+
+        let trivial_list = Syntax::new_list(&arena, "(", vec![], vec![atom], ")", vec![]);
+
+        match trivial_list {
+            List { children, .. } => {
+                assert_eq!(children.len(), 0);
+            }
+            Atom { .. } => unreachable!(),
+        }
+    }
+
     #[test]
     fn test_multiline_comment_ignores_leading_whitespace() {
         let pos = vec![SingleLineSpan {
