@@ -59,6 +59,7 @@ extern "C" {
     fn tree_sitter_json() -> ts::Language;
     fn tree_sitter_ocaml() -> ts::Language;
     fn tree_sitter_ocaml_interface() -> ts::Language;
+    fn tree_sitter_php() -> ts::Language;
     fn tree_sitter_python() -> ts::Language;
     fn tree_sitter_ruby() -> ts::Language;
     fn tree_sitter_rust() -> ts::Language;
@@ -122,7 +123,7 @@ pub fn from_language(language: guess::Language) -> TreeSitterConfig {
                 atom_nodes: (vec!["string_literal", "char_literal"])
                     .into_iter()
                     .collect(),
-                delimiter_tokens: (vec![("(", ")"), ("{", "}"), ("[", "]")]),
+                delimiter_tokens: (vec![("(", ")"), ("{", "}"), ("[", "]"), ("<", ">")]),
                 highlight_query: ts::Query::new(
                     language,
                     concat!(
@@ -333,6 +334,20 @@ pub fn from_language(language: guess::Language) -> TreeSitterConfig {
                 .unwrap(),
             }
         }
+        Php => {
+            let language = unsafe { tree_sitter_php() };
+            TreeSitterConfig {
+                name: "PHP",
+                language,
+                atom_nodes: (vec!["string"]).into_iter().collect(),
+                delimiter_tokens: (vec![("(", ")"), ("[", "]"), ("{", "}")]),
+                highlight_query: ts::Query::new(
+                    language,
+                    include_str!("../vendor/highlights/php.scm"),
+                )
+                .unwrap(),
+            }
+        }
         Python => {
             let language = unsafe { tree_sitter_python() };
             TreeSitterConfig {
@@ -379,7 +394,13 @@ pub fn from_language(language: guess::Language) -> TreeSitterConfig {
                 atom_nodes: (vec!["char_literal", "string_literal"])
                     .into_iter()
                     .collect(),
-                delimiter_tokens: (vec![("{", "}"), ("(", ")"), ("[", "]"), ("|", "|")]),
+                delimiter_tokens: (vec![
+                    ("{", "}"),
+                    ("(", ")"),
+                    ("[", "]"),
+                    ("|", "|"),
+                    ("<", ">"),
+                ]),
                 highlight_query: ts::Query::new(
                     language,
                     include_str!("../vendor/highlights/rust.scm"),
